@@ -243,6 +243,10 @@ class DependencyGraph(object):
         self.CUSTOMIZED_PRUNERS = {}
         self.IGNORED_LAYERS = []
 
+        # cache
+        self._in_channel_pruning_fn = set([p.prune_in_channels for p in self.REGISTERED_PRUNERS.values() if p is not None] + [p.prune_in_channels for p in self.CUSTOMIZED_PRUNERS.values() if p is not None])
+        self._out_channel_pruning_fn = set([p.prune_out_channels for p in self.REGISTERED_PRUNERS.values() if p is not None] + [p.prune_out_channels for p in self.CUSTOMIZED_PRUNERS.values() if p is not None])
+
     def build_dependency(
         self,
         model: torch.nn.Module,
@@ -270,7 +274,20 @@ class DependencyGraph(object):
         self.model = model
         self._module2name = {module: name for (
             name, module) in model.named_modules()}
-        
+
+        # Register customized pruners
+        if customized_pruners is not None:
+            for customized_module, customized_pruner in customized_pruners.items():
+                self.register_customized_layer(
+                    customized_module, customized_pruner)
+
+        # Ignore all sub-modules of customized layers
+        for layer_type in self.CUSTOMIZED_PRUNERS.keys():
+            for m in self.model.modules():
+                if isinstance(m, layer_type):
+                    for sub_module in m.modules():
+                        if sub_module != m:
+                            self.IGNORED_LAYERS.append(sub_module)
 
         # Detect unwrapped nn.parameters
         wrapped_parameters = []
@@ -294,25 +311,17 @@ class DependencyGraph(object):
             unwrapped_parameters = []
 
         unwrapped_detected = list( set(unwrapped_detected) - set([p for (p, _) in unwrapped_parameters]) )
-        if len(unwrapped_detected)>0:
-            warnings.warn("Unwrapped parameters detected: {}.\n\n DepGraph requires an unwrapped_parameter list to determine the pruning_dim of nn.Parameters. By default, the last dim of a parameter matrix will be pruned.".format([param_to_name[p] for p in unwrapped_detected]))
+        if len(unwrapped_detected)>0 and self.verbose:
+            warnings.warn("Unwrapped parameters detected: {}.\n Torch-Pruning will prune the last non-singleton dimension of a parameter. If you wish to customize this behavior, please provide an unwrapped_parameters argument.".format([param_to_name[p] for p in unwrapped_detected]))
         for p in unwrapped_detected:
-            unwrapped_parameters.append( UnwrappedParameters(parameters=p, pruning_dim=-1) ) # prune the last dim by daufault
+            # get the last dimension that >1
+            def last_non_singleton_dim(tensor):
+                non_singleton_dims = [i for i, s in enumerate(tensor.shape) if s > 1]
+                return non_singleton_dims[-1] if non_singleton_dims else None
+            pruning_dim = last_non_singleton_dim(p)
+            if pruning_dim is not None:
+                unwrapped_parameters.append( UnwrappedParameters(parameters=p, pruning_dim=pruning_dim) ) # prune the last non-singleton dim by daufault
         self.unwrapped_parameters = unwrapped_parameters
-        # Register customized pruners
-        if customized_pruners is not None:
-            for customized_module, customized_pruner in self.customized_pruners.items():
-                self.register_customized_layer(
-                    customized_module, customized_pruner)
-
-        # Ignore all sub-modules of customized layers
-        for layer_type in self.CUSTOMIZED_PRUNERS.keys():
-            for m in self.model.modules():
-                if isinstance(m, layer_type):
-                    for sub_module in m.modules():
-                        if sub_module != m:
-                            self.IGNORED_LAYERS.append(sub_module)
-
         # Build computational graph by tracing.
         self.module2node = self._trace(
             model, example_inputs, forward_fn, output_transform=output_transform
@@ -336,6 +345,9 @@ class DependencyGraph(object):
             pruner (tp.pruner.BasePruningFunc): a pruner for the specified layer type.
         """
         self.CUSTOMIZED_PRUNERS[layer_type] = layer_pruner
+        # Update cache
+        self._in_channel_pruning_fn = set([p.prune_in_channels for p in self.REGISTERED_PRUNERS.values() if p is not None] + [p.prune_in_channels for p in self.CUSTOMIZED_PRUNERS.values() if p is not None])
+        self._out_channel_pruning_fn = set([p.prune_out_channels for p in self.REGISTERED_PRUNERS.values() if p is not None] + [p.prune_out_channels for p in self.CUSTOMIZED_PRUNERS.values() if p is not None])
 
     def check_pruning_group(self, group: Group) -> bool:
         """check the group to avoid over-pruning. Return True if there are sufficient prunable elements.
@@ -344,23 +356,26 @@ class DependencyGraph(object):
         """
 
         for dep, idxs in group:
-            if function.is_out_channel_pruner(dep.handler):
+            if self.is_out_channel_pruning_fn(dep.handler):
                 prunable_chs = self.get_out_channels(
                     dep.target.module)
+                if prunable_chs is None: continue
                 if prunable_chs <= len(idxs):
                     return False
-            if function.is_in_channel_pruner(dep.handler):
+
+            if self.is_in_channel_pruning_fn(dep.handler):
                 prunable_in_chs = self.get_in_channels(
                     dep.target.module)
+                if prunable_in_chs is None: continue
                 if prunable_in_chs <= len(idxs):
                     return False
         return True
 
-    def is_out_channel_pruner(self, pruner: function.BasePruningFunc) -> bool:
-        return function.is_out_channel_pruner(pruner)
-
-    def is_in_channel_pruner(self, pruner: function.BasePruningFunc) -> bool:
-        return function.is_in_channel_pruner(pruner)
+    def is_out_channel_pruning_fn(self, fn: typing.Callable) -> bool:
+        return (fn in self._out_channel_pruning_fn)
+    
+    def is_in_channel_pruning_fn(self, fn: typing.Callable) -> bool:
+        return (fn in self._in_channel_pruning_fn)
 
     def get_pruning_plan(self, module: nn.Module, pruning_fn: typing.Callable, idxs: typing.Union[list, tuple]) -> Group:
         """ An alias of DependencyGraph.get_pruning_group for compatibility.
@@ -440,7 +455,7 @@ class DependencyGraph(object):
             if not isinstance(m, tuple(root_module_types)):
                 continue
 
-            pruner = self.REGISTERED_PRUNERS.get(ops.module2type(m), None)
+            pruner = self.get_pruner_of_module(m)
             if pruner is None or pruner.get_out_channels(m) is None:
                 continue
 
@@ -454,22 +469,28 @@ class DependencyGraph(object):
             for dep, _ in group:
                 module = dep.target.module
                 pruning_fn = dep.handler
-                if function.is_out_channel_pruner(pruning_fn):
+                if self.is_out_channel_pruning_fn(pruning_fn):
                     visited_layers.append(module)
                     if module in ignored_layers:
                         prunable_group = False
             if prunable_group:
                 yield group
 
-    def get_module_pruner(self, module):
-        return self.REGISTERED_PRUNERS.get(ops.module2type(module), None)
+    def get_pruner_of_module(self, module):
+        p = self.CUSTOMIZED_PRUNERS.get(module.__class__, None)
+        if p is None:
+            p = self.REGISTERED_PRUNERS.get(ops.module2type(module), None)
+        return p
 
     def get_out_channels(self, module_or_node):
         if isinstance(module_or_node, Node):
             module = module_or_node.module
+            pruning_dim = module_or_node.pruning_dim
         else:
             module = module_or_node
-        p = function.PrunerBox.get(ops.module2type(module), None)
+            pruning_dim = self.module2node[module].pruning_dim
+        p = self.get_pruner_of_module(module)
+        p.pruning_dim = pruning_dim
         if p is None:
             return None
         return p.get_out_channels(module)
@@ -477,9 +498,12 @@ class DependencyGraph(object):
     def get_in_channels(self, module_or_node):
         if isinstance(module_or_node, Node):
             module = module_or_node.module
+            pruning_dim = module_or_node.pruning_dim
         else:
             module = module_or_node
-        p = function.PrunerBox.get(ops.module2type(module), None)
+            pruning_dim = self.module2node[module].pruning_dim
+        p = self.get_pruner_of_module(module)
+        p.pruning_dim = pruning_dim
         if p is None:
             return None
         return p.get_in_channels(module)
@@ -527,32 +551,16 @@ class DependencyGraph(object):
             # Rule 1) - Inter-layer Dependency
             ###########################################
             for in_node in node.inputs:
-                handler = self.REGISTERED_PRUNERS.get(in_node.type)
-                if handler is None:
-                    handler = self.CUSTOMIZED_PRUNERS[in_node.class_type]
-                handler = handler.prune_out_channels
-
-                trigger = self.REGISTERED_PRUNERS.get(node.type)
-                if trigger is None:
-                    trigger = self.CUSTOMIZED_PRUNERS[node.class_type]
-                trigger = trigger.prune_in_channels
-
+                handler = self.get_pruner_of_module(in_node.module).prune_out_channels
+                trigger = self.get_pruner_of_module(node.module).prune_in_channels
                 dep = Dependency(
                     trigger=trigger, handler=handler, source=node, target=in_node
                 )
                 node.dependencies.append(dep)
 
             for out_node in node.outputs:
-                trigger = self.REGISTERED_PRUNERS.get(node.type)
-                if trigger is None:
-                    trigger = self.CUSTOMIZED_PRUNERS[node.class_type]
-                trigger = trigger.prune_out_channels
-
-                handler = self.REGISTERED_PRUNERS.get(out_node.type)
-                if handler is None:
-                    handler = self.CUSTOMIZED_PRUNERS[out_node.class_type]
-                handler = handler.prune_in_channels
-
+                trigger = self.get_pruner_of_module(node.module).prune_out_channels
+                handler = self.get_pruner_of_module(out_node.module).prune_in_channels
                 dep = Dependency(
                     trigger=trigger, handler=handler, source=node, target=out_node
                 )
@@ -571,14 +579,14 @@ class DependencyGraph(object):
         model.eval()
         gradfn2module = {}
         visited = {}
-        self._2d_4d = True # CNNs or Transformer, only for pytorch<=1.8
+        self._2d_4d = True # only for pytorch<=1.8
         def _record_grad_fn(module, inputs, outputs):
             if module not in visited:
                 visited[module] = 1
             else:
                 visited[module] += 1
             
-            if isinstance(module, ops.TORCH_LINEAR) and len(inputs[0].shape)==3:
+            if isinstance(module, nn.Linear) and len(outputs.shape)==3:
                 self._2d_4d=False
 
             if isinstance(outputs, tuple):
@@ -709,7 +717,7 @@ class DependencyGraph(object):
                                 if f[0].variable is p:
                                     is_unwrapped_param = True
                                     gradfn2module[f[0]] = p
-                                    self._module2name[p] = "UnwrappedParameter_{}".format(j)
+                                    self._module2name[p] = "UnwrappedParameter_{} ({})".format(j, p.shape)
                             if not is_unwrapped_param:
                                 continue
                         input_node = create_node_if_not_exists(f[0])
@@ -757,7 +765,6 @@ class DependencyGraph(object):
                         dep.index_mapping = _helpers._FlattenIndexMapping(
                             stride=stride, reverse=True
                         )
-
                 for dep in in_node.dependencies:
                     if dep.target == fc_node:
                         dep.index_mapping = _helpers._FlattenIndexMapping(
@@ -766,30 +773,37 @@ class DependencyGraph(object):
 
     def _update_reshape_index_mapping(self, reshape_node: Node):
         
-        in_channels = None
-        for n in reshape_node.inputs:
-            in_channels = self._infer_out_channels_recursively(n)
-            if in_channels is not None:  # =0 if there is a residual connection to model inputs
-                break
-            
+        # Only Supports 2D/4D tensors
+        # TODO: Better support for reshape/view/flatten
+        if hasattr(reshape_node.grad_fn, '_saved_self_sizes'): 
+            size = reshape_node.grad_fn._saved_self_sizes
+            if (len(size)!=1 and len(size)!=4):
+                return
+        else: # old pytorch versions
+            if not self._2d_4d:
+                return 
+
         out_channels = None
         for n in reshape_node.outputs:
             out_channels = self._infer_in_channels_recursively(n)
             if out_channels is not None:  # =0 if there is a residual connection to model inputs
                 break
         
+        in_channels = None
+        for n in reshape_node.inputs:
+            in_channels = self._infer_out_channels_recursively(n)
+            if in_channels is not None:  # =0 if there is a residual connection to model inputs
+                break
+        
         if out_channels is None or in_channels is None: return
         if out_channels==in_channels: return
-
-        # Only Supports 2D/4D tensors
-        if hasattr(reshape_node.grad_fn, '_saved_self_sizes'): 
-            if len(reshape_node.grad_fn._saved_self_sizes)!=1 and len(reshape_node.grad_fn._saved_self_sizes)!=4:
+        
+        if hasattr(reshape_node.grad_fn, '_saved_self_sizes'):
+            if len(size)==4 and size[1]*size[2]*size[3]!=out_channels:
                 return
-        else: # old pytorch versions
-            if not self._2d_4d:
-                return 
-
+        
         # Flatten
+        #print(reshape_node.grad_fn._saved_self_sizes, in_channels, out_channels)
         if out_channels > in_channels:
              for in_node in reshape_node.inputs:
                 for dep in reshape_node.dependencies:
